#ifdef TARGET_ZCODE;
struct long {
  int hi;
  int lo;
};

Global L_Zero : long = (0, 0);
Global L_One : long = (0, 1);
Global L_MinusOne : long = (-1, -1);
Global L_Billion : long = ($3b9a, $ca00);
Global L_Million : long = ($000f, $4240);
Global L_Thousand : long = (0, 1000);
Global L_Ten : long = (0, 10);

Constant SIGNED_MAX = $7fff;
Constant TOP_BIT = $8000;
#endif;
#ifdef TARGET_GLULX;
struct long {
  int long_val;
};

Global L_Zero : long = (0);
Global L_One : long = (1);
Global L_MinusOne : long = (-1);
Global L_Billion : long = (1000000000);
Global L_Million : long = (1000000);
Global L_Thousand : long = (1000);
Global L_Ten : long = (10);

Constant SIGNED_MAX = $7fffffff;
Constant TOP_BIT = $80000000;
#endif;

Declare LongAdd: in long a, in long b, return long;
Declare LongSub: in long a, in long b, return long;
Declare LongDouble: in/out long num;
Declare LongHalve: in/out long num;
Declare LongUnsignedMul: in long a, in long b, return long;
Declare LongSignedMul: in long a, in long b, return long;
Declare LongUnsignedDivMod: in long a, in long b, out long quo, out long rem;
Declare LongUnsignedDiv: in long a, in long b, return long;
Declare LongUnsignedMod: in long a, in long b, return long;
Declare LongSignedDivMod: in long a, in long b, out long quo, out long rem;
Declare LongSignedDiv: in long a, in long b, return long;
Declare LongSignedMod: in long a, in long b, return long;
Declare LongSqrt: in long num, return long;
Declare LongSignedCmp: in long a, in long b;
Declare LongUnsignedCmp: in long a, in long b;
Declare LongNeg: in/out long x;
Declare LongIsZero: in long x;
Declare LongSHR: in/out long num, in int t;
Declare LongNegIndex: in/out long a, in/out long b;

#ifdef TARGET_ZCODE;
[ NeedToCarry a b;
  if (a > 0 && b > 0) rfalse;
  if (a < 0 && b < 0) rtrue;
  if (a >= 0)
    return a + (b & SIGNED_MAX) < 0;
  else
    return b + (a & SIGNED_MAX) < 0;
];

[ LongAdd:
  in long a, in long b,
  local long sum,
  return long;
  
  sum.lo = a.lo + b.lo;
  sum.hi = a.hi + b.hi;
  if (NeedToCarry(a.lo, b.lo))
    (sum.hi)++;
  return sum;
];

[ LongSub:
  in long a, in long b,
  return long;
  
  LongNeg(b);
  return LongAdd(a, b);
];

! returns the bit lost off the top
[ LongDouble:
  in/out long num,
  local int c, local int t;
  
  c = (num.lo & TOP_BIT) ~= 0;
  t = (num.hi & TOP_BIT) ~= 0;
  num.lo = num.lo * 2;
  num.hi = num.hi * 2 | c;
  return t;
];

[ LongHalve:
  in/out long num,
  local int c;
  
  if (num.hi == -1 && num.lo == -1) {
    ! special case
    num.hi = 0;
    num.lo = 0;
    return;
  }
  
  c = num.hi & 1;
  num.hi = num.hi / 2;
  num.lo = num.lo / 2;
  if (c) { num.lo = num.lo | TOP_BIT; }
  else { num.lo = num.lo &~ TOP_BIT; }
];

! unsigned version of LongHalve
! the top bit is set to t, and the bit lost off the bottom is returned
[ LongSHR:
  in/out long num,
  in int t,
  local int c;

  c = num.hi & 1;
  num.hi = num.hi / 2;
  if (t) { num.hi = num.hi | TOP_BIT; }
  t = num.lo & 1;
  num.lo = num.lo / 2;
  if (c) { num.lo = num.lo | TOP_BIT; }
  else { num.lo = num.lo &~ TOP_BIT; }
  return t;
];

[ LongUnsignedMul:
  in long a, in long b,
  local long product,
  return long;
  
  ! I'm just a peasant...
  product.hi = product.lo = 0;
  
  while (~~LongIsZero(a)) {
    if (a.lo & 1) {
      product = LongAdd(product, b);
    }
    
    LongHalve(a);
    LongDouble(b);
  }
  return product;
];

[ LongSignedMul:
  in long a, in long b,
  local int n,
  local long product,
  return long;
  
  n = LongNegIndex(a, b);
  product = LongUnsignedMul(a, b);
  if (n) { LongNeg(product); }
  return product;
];

#ifndef UnsignedCompare;
[ UnsignedCompare x y u v;
    if (x == y) return 0;
    if (x < 0 && y >= 0) return 1;
    if (x >= 0 && y < 0) return -1;
    u = x&~TOP_BIT; v= y&~TOP_BIT;
    if (u > v) return 1;
    return -1;
];
#endif;

[ LongUnsignedDivMod:
  in long a, in long b,
  out long quo, out long rem,
  local long temp1,
  local long temp2,
  local int t;
  
  if (LongIsZero(b)) { t=0; t=t/0; }
  
  temp1 = b;
  rem = a;
  quo.hi = quo.lo = 0;
  temp2.hi = 0; temp2.lo = 1;
  
  while (LongUnsignedCmp(temp1, rem) <= 0) {
    t = LongDouble(temp2);
    if (t) break;
    LongDouble(temp1);
  }
  for (::) {
    if (LongSHR(temp2, t)) break;
    t = 0;
    LongSHR(temp1);
    if (LongUnsignedCmp(temp1, rem) <= 0) {
      rem = LongSub(rem, temp1);
      quo = LongAdd(quo, temp2);
    }
  }
];

[ LongUnsignedDiv:
  in long a, in long b,
  local long dummy,
  local long quo,
  return long;
  
  LongUnsignedDivMod(a, b, quo, dummy);
  return quo;
];

[ LongUnsignedMod:
  in long a, in long b,
  local long dummy,
  local long rem,
  return long;
  
  LongUnsignedDivMod(a, b, dummy, rem);
  return rem;
];

[ LongSignedDivMod:
  in long a, in long b,
  out long quo,
  out long rem,
  local int n;
  
  n = LongNegIndex(a, b);
  LongUnsignedDivMod(a, b, quo, rem);
  if (n) { LongNeg(quo); LongNeg(rem); }
];

[ LongSignedDiv:
  in long a, in long b,
  local long dummy,
  local int n,
  local long quo,
  return long;
  
  n = LongNegIndex(a, b);
  LongUnsignedDivMod(a, b, quo, dummy);
  if (n) { LongNeg(quo); }
  return quo;
];

[ LongSignedMod:
  in long a, in long b,
  local long dummy,
  local int n,
  local long rem,
  return long;
  
  n = LongNegIndex(a, b);
  LongUnsignedDivMod(a, b, dummy, rem);
  if (n) { LongNeg(rem); }
  return rem;
];

! c = abs(a); d = abs(b); rtrue if a*b is negative
[ LongNegIndex:
  in/out long a, in/out long b,
  local int rv;

  rv = ((a.hi & TOP_BIT) ~= (b.hi & TOP_BIT));
  
  if (a.hi & TOP_BIT) { LongNeg(a); }
  if (b.hi & TOP_BIT) { LongNeg(b); }
  
  return rv;
];

[ LongSqrt:
  in long num,
  local long orig,
  local long diff,
  local long last,
  local long result,
  return long;
  
  orig = num;
  LongHalve(num);
  result = num;
  if (result.hi & TOP_BIT) { return L_MinusOne; }
  
  if (LongIsZero(result)) { return; }
  for (::) {
    last = result;
    diff = result;
    LongHalve(diff);
    diff = LongSub(diff, LongUnsignedDiv(num, result));
    if (LongIsZero(diff)) { break; }
    
    result = LongSub(result, diff);
    
    if (LongUnsignedCmp(LongAdd(last, L_One), result) == 0) { break; }
  }
  return result;
];  

[ LongSignedCmp:
  in long a, in long b;
  
  if (a.hi ~= b.hi) {
    return a.hi - b.hi;
  }
  return a.lo - b.lo;
];

[ LongUnsignedCmp:
  in long a, in long b;
  
  if (a.hi ~= b.hi) {
    return UnsignedCompare(a.hi, b.hi);
  }
  return UnsignedCompare(a.lo, b.lo);
];

[ LongNeg:
  in/out long x;
  
  x.hi = ~(x.hi);
  x.lo = ~(x.lo);
  x = LongAdd(x, L_One);
];

[ LongIsZero:
  in long x;
  
  if (x.hi == 0 && x.lo == 0) rtrue;
  rfalse;
];

[ MakeLong:
  in int a, in int b, in int c, in int d,
  local long result,
  local long segment,
  local int n,
  return long;
  
  if (a < 0) { n = 1; a = -a; }
  if (b < 0) { n = 1; b = -b; }
  if (c < 0) { n = 1; c = -c; }
  if (d < 0) { n = 1; d = -d; }
  
  result.lo = d;
  if (c) {
    segment.lo = c;
    result = LongAdd(result, LongUnsignedMul(segment, L_Thousand));
  }
  if (b) {
    segment.lo = b;
    result = LongAdd(result, LongUnsignedMul(segment, L_Million));
  }
  if (a) {
    segment.lo = a;
    result = LongAdd(result, LongUnsignedMul(segment, L_Billion));
  }

  if (n) LongNeg(result);
  return result;
];

[ DumpLong:
  in long x;
  
  print "Hi=", x.hi, " Lo=", x.lo;
];

Array printlong_digits -> 10;

[ PrintLong:
  in long x, in int commas,
  local long rem,
  local int negative,
  local int digit;
  
  negative = digit = 0;
  if (x.hi & TOP_BIT) {
    negative = 1;
    LongNeg(x);
  }
  
  while (~~LongIsZero(x)) {
    LongUnsignedDivMod(x, L_Ten, x, rem);
    printlong_digits->(digit++) = '0' + rem.lo;
  }
  
  if (digit == 0) {
    print "0";
    return;
  }
  
  if (negative) print "-";
  
  while (digit > 0) {
    print (char) printlong_digits->(--digit);
    if (commas && (digit >= 3) && (digit % 3 == 0))
      print ",";
  }
];

[ ParseLong:
  in int buffer, in int length,
  local long result,
  local long digit,
  local int n,
  local int i,
  local int c,
  return long;
  
  for (i=0: i<length: i++) {
    c = buffer->i;
    if (i == 0 && c == '-') {
      n = 1;
    } else if (i >= '0' && i <= '9') {
      digit.lo = i - '0';
      result = LongUnsignedMul(result, L_Ten);
      result = LongAdd(result, digit);
    } else {
      break;
    }
  }
  if (n) LongNeg(result);
  return result;
];
#endif;  ! TARGET_ZCODE

#ifdef TARGET_GLULX;
[ LongAdd:
  in long a, in long b,
  local long sum,
  return long;
  
  sum.long_val = a.long_val + b.long_val;
  return sum;
];

[ LongSub:
  in long a, in long b,
  local long diff,
  return long;
  
  diff.long_val = a.long_val - b.long_val;
  return diff;
];

! returns the bit lost off the top
[ LongDouble:
  in/out long num,
  local int t;
  
  t = (num.long_val & TOP_BIT) ~= 0;
  num.long_val = num.long_val * 2;
  return t;
];

[ LongHalve:
  in/out long num;
  
  num.long_val = num.long_val / 2;
];

! unsigned version of LongHalve
! the top bit is set to t, and the bit lost off the bottom is returned
[ LongSHR:
  in/out long num,
  in int t,
  local int c, local int v;

  c = num.long_val & 1;
  v = num.long_val / 2;
  if (t) v = v | TOP_BIT;
  else v = v &~ TOP_BIT;
  num.long_val = v;
  return c;
];

! all long math is signed on Glulx
[ LongUnsignedMul:
  in long a, in long b,
  local long product,
  return long;
  
  product.long_val = a.long_val * b.long_val;
  return product;
];

[ LongSignedMul:
  in long a, in long b,
  local long product,
  return long;
  
  product.long_val = a.long_val * b.long_val;
  return product;
];

[ LongUnsignedDivMod:
  in long a, in long b,
  out long quo, out long rem;
  
  quo.long_val = a.long_val / b.long_val;
  rem.long_val = a.long_val % b.long_val;
];

[ LongUnsignedDiv:
  in long a, in long b,
  local long quo,
  return long;
  
  quo.long_val = a.long_val / b.long_val;
  return quo;
];

[ LongUnsignedMod:
  in long a, in long b,
  local long rem,
  return long;
  
  rem.long_val = a.long_val % b.long_val;
  return rem;
];

[ LongSignedDivMod:
  in long a, in long b,
  out long quo,
  out long rem;
  
  quo.long_val = a.long_val / b.long_val;
  rem.long_val = a.long_val % b.long_val;
];

[ LongSignedDiv:
  in long a, in long b,
  local long quo,
  return long;
  
  quo.long_val = a.long_val / b.long_val;
  return quo;
];

[ LongSignedMod:
  in long a, in long b,
  local long rem,
  return long;
  
  rem.long_val = a.long_val % b.long_val;
  return rem;
];

! c = abs(a); d = abs(b); rtrue if a*b is negative
[ LongNegIndex:
  in/out long a, in/out long b,
  local int rv;

  rv = ((a.long_val & TOP_BIT) ~= (b.long_val & TOP_BIT));
  
  if (a.long_val < 0) { a.long_val = -(a.long_val); }
  if (b.long_val < 0) { b.long_val = -(b.long_val); }
  
  return rv;
];

[ LongSqrt:
  in long num,
  local int num0,
  local int orig,
  local int diff,
  local int last,
  local int result0,
  local long result,
  return long;
  
  orig = num.long_val;
  if (orig < 0) { return L_MinusOne; }
  result0 = num0 = orig / 2;
  if (result0 == 0) { return L_Zero; }
  
  for (::) {
    last = result0;
    diff = result0 / 2;
    diff = diff - (num0 / result0);
    if (diff == 0) { break; }
    
    result0 = result0 - diff;
    
    if (last + 1 == result0) { break; }
  }
  result.long_val = result0;
  return result;
];  

[ LongSignedCmp:
  in long a, in long b;
  
  return a.long_val - b.long_val;
];

! even this is signed on Glulx...
[ LongUnsignedCmp:
  in long a, in long b;
  
  return a.long_val - b.long_val;
];

[ LongNeg:
  in/out long x;
  
  x.long_val = -(x.long_val);
];

[ LongIsZero:
  in long x;
  
  if (x.long_val == 0) rtrue;
  rfalse;
];

[ MakeLong:
  in int a, in int b, in int c, in int d,
  local long result,
  local int n,
  local int x,
  return long;
  
  if (a < 0) { n = 1; a = -a; }
  if (b < 0) { n = 1; b = -b; }
  if (c < 0) { n = 1; c = -c; }
  if (d < 0) { n = 1; d = -d; }
  
  x = (a * 1000000000) + (b * 1000000) + (c * 1000) + d;
  if (n) { x = -x; }
  
  result.long_val = x;
  return result;
];

[ DumpLong:
  in long x;
  
  print "LV=", x.long_val;
];

Array printlong_digits -> 10;

[ PrintLong:
  in long x, in int commas,
  local int x0,
  local int rem,
  local int negative,
  local int digit;
  
  negative = digit = 0;
  x0 = x.long_val;
  if (x0 < 0) {
    negative = 1;
    x0 = -x0;
  }
  
  while (x0) {
    rem = x0 % 10;
    x0 = x0 / 10;
    printlong_digits->(digit++) = '0' + rem;
  }
  
  if (digit == 0) {
    print "0";
    return;
  }
  
  if (negative) print "-";
  
  while (digit > 0) {
    print (char) printlong_digits->(--digit);
    if (commas && (digit >= 3) && (digit % 3 == 0))
      print ",";
  }
];

[ ParseLong:
  in int buffer, in int length,
  local int result0,
  local int n,
  local int i,
  local int c,
  local long result,
  return long;
  
  for (i=0: i<length: i++) {
    c = buffer->i;
    if (i == 0 && c == '-') {
      n = 1;
    } else if (i >= '0' && i <= '9') {
      result0 = result0 * 10 + i - '0';
    } else {
      break;
    }
  }
  if (n) { result0 = -result0; }
  result.long_val = result0;
  return result;
];
#endif; ! TARGET_GLULX
