Trace frames on;

struct DSTR {
  int length;
  int charptr;
  destructor [; mfree(self->charptr); ];
};

#ifdef TARGET_ZCODE;
Array measure_buffer buffer 100;
#ifnot;
Global measured_chars;
[ MeasureStringFilter; measured_chars++; ];
[ CopyStringFilter ch; (measured_chars++)->0 = ch; ];
#endif;

Declare NewDSTR: in int str, return DSTR*;
Declare StrCat: in DSTR* s1, in DSTR* s2, return DSTR*;
Declare StrChr: in DSTR* s, in int c, return int;
Declare StrCmp: in DSTR* s1, in DSTR* s2, return int;
Declare StrUpper: in DSTR* s, return DSTR*;
Declare StrLower: in DSTR* s, return DSTR*;
Declare StrLen: in DSTR* s, return int;
Declare SubStr: in DSTR* s, in int start, in int len, return DSTR*;

[ NewDSTR:
  return DSTR*,
  in int zstring,
  local DSTR* result,
  local int i,
  local int len;
  
  #ifdef TARGET_ZCODE;
  @output_stream 3 measure_buffer;
  print (string) zstring;
  @output_stream -3;
  len = measure_buffer-->0;
  if (~~len) rfalse;
  #ifnot;
  @getiosys sp sp;
  @setiosys 1 MeasureStringFilter;
  measured_chars = 0;
  print (string) zstring;
  len = measured_chars;
  if (~~len) { @stkswap; @setiosys sp sp; rfalse; }
  #endif;
  
  result = new DSTR;
  result->length = len;
  result->charptr = malloc(len);
  #ifdef TARGET_ZCODE;
  for (i=0: i<len: i++)
    result->charptr->i = measure_buffer->(2+i);
  #ifnot;
  @setiosys 1 CopyStringFilter;
  measured_chars = result->charptr;
  print (string) zstring;
  @stkswap; @setiosys sp sp;
  #endif;
  
  return result;
];

[ StrCat:
  return DSTR*,
  in DSTR* s1,
  in DSTR* s2,
  local DSTR* result,
  local int i,
  local int j;
  
  if (~~s1) return s2;
  if (~~s2) return s1;
  
  result = new DSTR;
  result->length = s1->length + s2->length;
  result->charptr = malloc(result->length);
  for (i=0,j=0: i<s1->length: i++, j++)
    result->charptr->j = s1->charptr->i;
  for (i=0: i<s2->length: i++, j++)
    result->charptr->j = s2->charptr->i;
  
  return result;
];

[ StrUpper:
  in DSTR* str,
  local DSTR* result,
  local int i,
  local int c,
  return DSTR*;
  
  if (~~str) rfalse;
  
  result = new DSTR;
  i = str->length;
  result->length = i;
  result->charptr = malloc(i);
  for (i--: i>=0: i--) {
    c = str->charptr->i;
    if (c >= 'a' && c <= 'z')
      c = c + ('A' - 'a');
    result->charptr->i = c;
  }
  return result;
];

[ StrLower:
  in DSTR* str,
  local DSTR* result,
  local int i,
  local int c,
  return DSTR*;
  
  if (~~str) rfalse;
  
  result = new DSTR;
  i = str->length;
  result->length = i;
  result->charptr = malloc(i);
  for (i--: i>=0: i--) {
    c = str->charptr->i;
    if (c >= 'A' && c <= 'Z')
      c = c + ('a' - 'A');
    result->charptr->i = c;
  }
  return result;
];

[ StrChr:
  in DSTR* str,
  in int ch,
  local int i;
  
  if (str)
    for (i=0: i<str->length: i++)
      if (str->charptr->i == ch)
        return i;
  
  return -1;
];

[ SubStr:
  in DSTR* str,
  in int start,
  in int len,
  local DSTR* result,
  return DSTR*;

  if ((~~str) || (start >= str->length))
    rfalse;
  if (start < 0) { len = len + start; start = 0; }
  if (start + len > str->length)
    len = str->length - start;
  if (~~len) rfalse;
  
  result = new DSTR;
  result->length = len;
  result->charptr = malloc(len);
  while (len-- > 0)
    result->charptr->len = str->charptr->(start+len);
  return result;
];

[ StrLen:
  in DSTR* str;
  
  if (~~str) rfalse;
  return str->length;
];

[ StrCmp:
  in DSTR* s1,
  in DSTR* s2,
  local int l1, local int l2,
  local int p1, local int p2,
  local int i,
  local int v;
  
  if (~~s1) {
    if (s2) return -1;
    return 0;
  }
  if (~~s2) return 1;
  
  l1 = s1->length; l2 = s2->length;
  p1 = s1->charptr; p2 = s2->charptr;
  for (i=0: i<l1 && i<l2: i++) {
    v = p1->i - p2->i;
    if (v) return v;
  }
  return l1 - l2;
];

[ StrICmp:
  in DSTR* s1,
  in DSTR* s2,
  local int l1, local int l2,
  local int p1, local int p2,
  local int i,
  local int v,
  local int c;
  
  if (~~s1) {
    if (s2) return -1;
    return 0;
  }
  if (~~s2) return 1;
  
  l1 = s1->length; l2 = s2->length;
  p1 = s1->charptr; p2 = s2->charptr;
  for (i=0: i<l1 && i<l2: i++) {
    c = p1->i;
    if (c >= 'A' && c <= 'Z') c = c + ('a' - 'A');
    v = p2->i;
    if (v >= 'A' && v <= 'Z') v = v + ('a' - 'A');
    v = c - v;
    if (v) return v;
  }
  return l1 - l2;
];

[ PrintDSTR:
  in DSTR* str,
  local int i,
  local int len;
  
  if (str) {
    len = str->length;
    for (i=0: i<len: i++)
      print (char) str->charptr->i;
  }
];
